from interlude import interact
from pkg_resources import resource_filename
from pprint import pprint

import doctest
import unittest
import tempfile
import tarfile

TEST_USBIDS_FILE = resource_filename(__name__, 'data/testusb.ids')
TEST_DATA = resource_filename(__name__, 'data/mocktree.tgz')

optionflags = doctest.NORMALIZE_WHITESPACE | \
              doctest.ELLIPSIS | \
              doctest.REPORT_ONLY_FIRST_FAILURE

TESTFILES = [
    'usbinfo.rst',
    'device.rst',
]

TEMPDIR = tempfile.mkdtemp()

tgz = tarfile.open(TEST_DATA)
tgz.extractall(TEMPDIR)

# define the testing root path
MOCK_SYS = TEMPDIR + '/sys/bus/usb/devices'


def test_suite():
    return unittest.TestSuite([
        doctest.DocFileSuite(
            filename,
            optionflags=optionflags,
            globs={'interact': interact,
                   'pprint': pprint,
                   'TEST_USBIDS_FILE': TEST_USBIDS_FILE,
                   'MOCK_SYS': MOCK_SYS,
                   },
        ) for filename in TESTFILES
    ])

if __name__ == '__main__':  # pragma NO COVERAGE
    unittest.main(defaultTest='test_suite')  # pragma NO COVERAGE



from pkg_resources import resource_filename
DEFAULT_USBIDS_FILE = resource_filename(__name__, 'data/usb.ids')

LISTOF_MAP = {
    '': '_usb_ids',
    'C': '_device_classes',
    'AT': '_audio_class_terminal_types',
    'HID': '_hid_descriptor_types',
    'R': '_hid_descriptor_item_types',
    'BIAS': '_physical_descriptor_bias_types',
    'PHY': '_physical_descriptor_item_types',
    'HUT': '_hid_usages',
    'L': '_languages',
    'HCC': '_country_codes',
    'VT': '_video_class_terminal_types',
}

BLOCK_WITH_1_LEVELS = [
    'AT',
    'HID',
    'R',
    'BIAS',
    'PHY',
    'HCC',
    'VT',
]

BLOCK_WITH_3_LEVELS = ['C']


class USBInfo(object):

    def __init__(self, usbid_file=DEFAULT_USBIDS_FILE):
        self._usbid_file = usbid_file
        for value in LISTOF_MAP.values():
            setattr(self, value, None)

    def _parse_usbids(self):
        """
        the parser, which reads the usbids file and returns the information
        """
        block_0 = None
        block_1 = None
        block_2 = None
        block_type_key = None

        try:
            with open(self._usbid_file, 'r') as raw:
                for line in raw:
                    line = line.strip('\n')
                    line = line.rstrip()  # whitespaces are causing errors
                    if not line or line[0] == '#':
                        continue

                    if not line[0] == '\t':
                        # lvl 0
                        left, right = line.split('  ', 1)

                        if not ' ' in left:
                            # check usb_ids
                            block_id = left
                            block_type_key = ''
                        else:
                            block_type_key, block_id = left.split(' ')

                        block_type = LISTOF_MAP[block_type_key]
                        block_0 = getattr(self, block_type)
                        if block_0 is None:
                            block_0 = {}
                            setattr(self, block_type, block_0)

                        # check if block has only 1 level. so that these blocks
                        # dont get filled with unnecessary empty dicts.
                        # because later on we check if a block has 3 levels,
                        # the blocks with 2 levels get automatically filled
                        # with the right  values
                        block_1 = {}
                        if block_type_key in BLOCK_WITH_1_LEVELS:
                            block_0[block_id] = (right)
                            continue
                        else:
                            block_0[block_id] = (right, block_1)
                            continue

                    if line.startswith('\t\t'):
                        # level 2
                        line = line.strip('\t\t')
                        left, right = line.split('  ', 1)
                        block_2[left] = right
                        continue

                    if not line.startswith('\t'):
                        raise ValueError("File corrupt")

                    # level 1
                    line = line.strip('\t')
                    left, right = line.split('  ', 1)
                    if block_type_key in BLOCK_WITH_3_LEVELS:
                        block_2 = {}
                        block_1[left] = (right, block_2)
                    else:
                        block_1[left] = right

        except IOError:
            raise IOError("Can't open file %s in order to parse usbinfo." %
                          self._usbid_file)

    @property
    def usb_ids(self):
        """
        dict of usbids with key as vendor id and value as a tuple
        consisting of vendor name and dict of key as device id and value
        as device name
        """
        if self._usb_ids is None:
            self._parse_usbids()
        return self._usb_ids

    @property
    def device_classes(self):
        """
        dict of device classes with key as class id and value as a tuple
        consisting of class name and dict of key as subclass id and value
        as a tuple consisting of subclass name and dict of key as protocol
        id and value as protocol name

        example::

            '09': ('Hub',
                {'00': ('Unused',
                    {'00': 'Full speed (or root) hub',
                     '01': 'Single TT',
                     '02': 'TT per port'})})
        """
        if self._device_classes is None:
            self._parse_usbids()
        return self._device_classes

    @property
    def audio_class_terminal_types(self):
        """
        dict of audio class terminal types with key as terminal type and
        value as terminal name
        """
        if self._audio_class_terminal_types is None:
            self._parse_usbids()
        return self._audio_class_terminal_types

    @property
    def hid_descriptor_types(self):
        """
        dict of HID descriptor types with key as HID descriptor type and
        value as descriptor type name
        """
        if self._hid_descriptor_types is None:
            self._parse_usbids()
        return self._hid_descriptor_types

    @property
    def hid_descriptor_item_types(self):
        """
        dict of HID descriptor item types with key as item type and
        value as item type name
        """
        if self._hid_descriptor_item_types is None:
            self._parse_usbids()
        return self._hid_descriptor_item_types

    @property
    def physical_descriptor_bias_types(self):
        """
        dict of BIAS item types with key as BIAS item type and
        value as item type name
        """
        if self._physical_descriptor_bias_types is None:
            self._parse_usbids()
        return self._physical_descriptor_bias_types

    @property
    def physical_descriptor_item_types(self):
        """
        dict of PHY item types with key as PHY item type and
        value as item type name
        """
        if self._physical_descriptor_item_types is None:
            self._parse_usbids()
        return self._physical_descriptor_item_types

    @property
    def hid_usages(self):
        """
        dict of HID usages with key as HID id and value as a tuple
        consisting of  HID usage page name and a dict of key as hid usage
        and value as hid usage name
        """
        if self._hid_usages is None:
            self._parse_usbids()
        return self._hid_usages

    @property
    def languages(self):
        """
        dict of languages with key as language id and value as a tuple
        consisting of language name and a dict of key as dialect id and
        value as dialect name
        """
        if self._languages is None:
            self._parse_usbids()
        return self._languages

    @property
    def country_codes(self):
        """
        dict of codes with key as code id and value as a tuple
        consisting of language name and a dict of key as dialect id and
        value as dialect name
        """
        if self._country_codes is None:
            self._parse_usbids()
        return self._country_codes

    @property
    def video_class_terminal_types(self):
        """
        dict of video class terminal types with key as terminal type and
        value as terminal type name
        """
        if self._video_class_terminal_types is None:
            self._parse_usbids()
        return self._video_class_terminal_types

# initialize as a singleton, so that the parser only runs once
USBINFO = USBInfo()



import os
import re
from usbid.usbinfo import USBINFO

# regex which are often needed
PARENT = re.compile("^\d{1}-{1}\d{1}$")
CHILD = re.compile("^[0-9\-\.]+$")
TTY = re.compile(".*:\d{1}.\d{1}")


def usb_roots(root_path='/sys/bus/usb/devices'):
    """
    This returns a dict of the usb roots.
    """
    # todo try except when access on wrong key
    usb_roots = {}
    for root in os.listdir(root_path):
        if root.startswith('usb'):
            root_id = int(root[3:])
            usb_roots[root_id] = DeviceNode(
                root_id,
                os.path.join(root_path, root),
                root_id,
                True
            )
    return usb_roots


def devicelist(root_path='/sys/bus/usb/devices'):
    """
    This returns a list of DeviceNode objects
    """
    res = []

    def walk(parent):
        for child in parent.values():
            res.append(child)
            walk(child)
    walk(usb_roots(root_path=root_path))
    return res


class DeviceNode(object):
    def __init__(self, own_id, fs_path, parent, is_root=False,
                 usbinfo=USBINFO):
        self.own_id = own_id
        self.fs_path = fs_path
        self.parent = parent
        self.is_root = is_root
        self.usbinfo = usbinfo

    @property
    def path(self):
        """
        returns path of the Devicenode as list of integers
        """
        current = self
        path = [self.own_id]
        while not current.is_root:
            current = current.parent
            path.insert(0, current.own_id)
        return path

    def keys(self):
        res = []
        for node in os.listdir(self.fs_path):
            if self.is_root and PARENT.match(node):
                node_r = node.rsplit("-", 1)[1]
                res.append(int(node_r))
            elif CHILD.match(node):
                node_r = node.rsplit(".", 1)[1]
                res.append(int(node_r))
        return res

    def values(self):
        return [_[1] for _ in self.items()]

    def items(self):
        res = []
        for node in os.listdir(self.fs_path):
            if self.is_root and PARENT.match(node):
                child_id = int(node.rsplit("-", 1)[1])
                dev = DeviceNode(child_id, os.path.join(self.fs_path, node),
                                 self.fs_path)
                res.append((child_id, dev))
            elif CHILD.match(node):
                child_id = int(node.rsplit(".", 1)[1])
                dev = DeviceNode(child_id, os.path.join(self.fs_path, node),
                                 self.fs_path)
                res.append((child_id, dev))
        return res

    def __getitem__(self, key):
        for node in os.listdir(self.fs_path):
            if PARENT.match(node):
                node_r = node.rsplit("-", 1)[1]
                if key == int(node_r):
                    dev = DeviceNode(key, os.path.join(self.fs_path, node),
                                     self.fs_path)
                    return dev

            if CHILD.match(node):
                node_r = node.rsplit(".", 1)[1]
                if key == int(node_r):
                    dev = DeviceNode(key, os.path.join(self.fs_path, node),
                                     self.fs_path)
                    return dev
        raise KeyError("No such Device with %s with in path %s" %
                       (key, '/'.join([str(_) for _ in self.path])))

    @property
    def idVendor(self):
        """
        get the vendor id as hex
        test if zerofill needed? res = res.zfill(4)
        """
        with open(os.path.join(self.fs_path, "idVendor"), "r") as fio:
            res = fio.read().strip("\n\x00")
        return res

    @property
    def idProduct(self):
        """
        get the product id as hex
        """
        with open(os.path.join(self.fs_path, "idProduct"), "r") as fio:
            res = fio.read().strip("\n\x00")
        return res

    @property
    def nameVendor(self):
        """
        get vendor name from USBINFO db file
        """
        try:
            return self.usbinfo.usb_ids[self.idVendor][0]
        except KeyError:
            return "UNKNOWN"

    @property
    def nameProduct(self):
        """
        get product name from USBINFO db file
        """
        try:
            return self.usbinfo.usb_ids[self.idVendor][1][self.idProduct]
        except KeyError:
            return "UNKNOWN"

    @property
    def tty(self):
        """
        return name of serial device or None if its not a serial device
        """
        for filename in os.listdir(self.fs_path):
            if TTY.match(filename):
                sub_path_name = os.path.join(self.fs_path, filename)
                for filename in os.listdir(sub_path_name):
                    if filename.startswith("tty"):
                        return filename

    def __str__(self):
        """
        return string with information nicely formatted
         """
        result = ""
        result += "idProduct: " + self.idProduct + "\n"
        result += "idVendor: " + self.idVendor + "\n"
        result += "Product Name: " + self.nameProduct + "\n"
        result += "Vendor Name: " + self.nameVendor
        return result

        # other useful stats which can be accessed:
        # port_number, address, bus, bDeviceClass,bDeviceProtocol,
        # bDeviceSubClass,bcdDevice,bcdUSB
